﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using Mol;

namespace BigExamples
{
    /// <summary>
    /// BigDecimal を使った例題集<br/>
    /// 
    /// 含まれているもの：Sqrt PI Sin Cos Atan E Log Exp　と線形連立方程式の解法（LuDecompとLuSolve）<br/>
    /// 例題ですので極力単純な形になっています。エラーチェックなども殆どしていません。<br/>
    /// このソースコードは自由に利用してかまいません。<br/>
    /// 
    /// Copyright (C) 2012 Shigeo Kobayashi. All rights reserved.
    /// 
    /// </summary>
    public class BigMath
    {
        /// <summary>
        /// 線形連立方程式 Ax = b (x と b は一次元のベクトル)を解く前処理として、正方行列 A を LU 分解します。
        /// A はは結果(LU)で上書きされます。 A や b の要素に null がある場合は値ゼロが仮定されます。
        /// </summary>
        /// <param name="A">係数行列／LU分解の結果</param>
        /// <param name="prec">必要な有効桁数</param>
        /// <returns>ピボット交換用のインデックス配列(後のLuSolve()で使用)</returns>
        public static int [] LuDecomp(BigDecimal [,] A,int prec)
        {
            int n = A.GetLength(0);
            BigDecimal zero = BigDecimal.Zero;
            BigDecimal one  = BigDecimal.One;
            BigDecimal r    = null;
            int [] ps     = new int[n];
            BigDecimal [] scales = new BigDecimal[n];

            // ピボット要素の選択（行単位）
            for(int i = 0; i<n ; ++i)
            {
                ps[i] = i;
                BigDecimal nrmrow  = zero;
                for(int j = 0;j< n;++j)
                {
                    if (A[i, j] == null) A[i, j] = BigDecimal.Zero;
                    BigDecimal v = A[i,j];
                    if(BigDecimal.CheckNumber(v)<0) v = -v;
                    if(v>nrmrow) nrmrow = v;
                }
                if (BigDecimal.IsZero(nrmrow)) throw new Exception("Singular matrix in LuDecomp(phase-1):"+i);
                scales[i] = BigDecimal.Div(null,ref r,one,nrmrow,prec);
            }

            // 消去法による三角化
            int n1     = n - 1;
            for(int k = 0;k<n1;++k)
            {
                BigDecimal  biggst = zero;
                int         pividx = 0;
                for (int i = k;i<n;++i)
                {
                    BigDecimal size = A[ps[i],k]*scales[ps[i]];
                    if(BigDecimal.CheckNumber(size)<0) BigDecimal.Negate(size);
                    if( size>biggst)
                    {
                        biggst = size;
                        pividx  = i;
                    }
                }

                if (BigDecimal.IsZero(biggst)) throw new Exception("Singular matrix in LuDecomp(phase-2):" + k);
                if (pividx!=k)
                {
                    int j = ps[k];
                    ps[k] = ps[pividx];
                    ps[pividx] = j;
                }
                BigDecimal pivot = A[ps[k],k];
                BigDecimal EM = null;
                for (int i = (k + 1); i < n; ++i)
                {
                    EM = BigDecimal.Div(EM, ref r, A[ps[i], k], pivot, prec);
                    A[ps[i], k] = EM.Clone();
                    BigDecimal mul = null;
                    for (int j = (k + 1); j < n; ++j)
                    {
                        mul = BigDecimal.Mul(mul, EM, A[ps[k], j], prec);
                        A[ps[i], j] -= mul;
                    }
                }
            }
            if (BigDecimal.IsZero(A[ps[n1], n1])) throw new Exception("Singular matrix in LuDecomp(phase-3)" + n1);
            return ps;
        }

        /// <summary>
        /// LuDecomp()で LU 分解された A を使用して線形連立式 Ax = b を解きます。
        /// 値の設定されていない b の要素（null）はゼロとして扱われます。
        /// </summary>
        /// <param name="A">LU 分解された係数行列</param>
        /// <param name="b">定数ベクトル</param>
        /// <param name="ps">ピボット変換用インデックス</param>
        /// <param name="prec">必要な有効桁数</param>
        /// <returns>解ベクトル x</returns>
        public static BigDecimal [] LuSolve(BigDecimal [,] A,BigDecimal [] b,int [] ps,int prec)
        {
            int n = ps.Length;
            BigDecimal zero = BigDecimal.Zero;
            BigDecimal [] x = new BigDecimal[n];
            BigDecimal r = null;
            for(int i = 0;i<n;++i)
            {
                BigDecimal dot = zero;
                BigDecimal mul = null;
                for(int j = 0;j<i;++j)
                {
                    mul  = BigDecimal.Mul(mul, A[ps[i], j], x[j], prec);
                    dot -= mul;
                }
                if (b[ps[i]] == null) x[i] = dot;
                else                  x[i] = b[ps[i]] + dot;
            }
            for(int i = n-1;i>=0;--i)
            {
                BigDecimal dot = zero;
                BigDecimal mul = null;
                for (int j = (i + 1); j < n; ++j)
                {
                    mul = BigDecimal.Mul(mul, A[ps[i], j], x[j], prec);
                    dot += mul;
                }
                dot = x[i]-dot;
                x[i] = BigDecimal.Div(null,ref r, dot, A[ps[i], i], prec);
            }
            return x;
        }

        /// <summary>
        /// BigDecimal 値 v の平方根を計算します。
        /// </summary>
        /// <param name="v">BigDecimal 値（&gt;=0）</param>
        /// <param name="prec">必要有効桁数</param>
        /// <returns>v の平方根</returns>
        public static BigDecimal Sqrt(BigDecimal v, int prec)
        {
            return BigDecimal.Sqrt(null,v, prec);
        }

        /// <summary>
        /// 円周率を(マチンの公式を使用して) prec 桁まで計算します。
        /// </summary>
        /// <param name="prec">必要桁数</param>
        /// <returns>円周率</returns>
        public static BigDecimal PI(int prec)
        {
            int double_fig    = BigDecimal.MaxDouble.DigitCount();
            int n             = prec * 2; 

            BigDecimal m25    = new BigDecimal("-0.04");
            BigDecimal m57121 = new BigDecimal("-57121");
            BigDecimal r      = null;
            BigDecimal pi     = BigDecimal.Zero;

            BigDecimal d      = BigDecimal.One;
            BigDecimal k      = BigDecimal.One;
            BigDecimal w      = BigDecimal.One;
            BigDecimal t      = new BigDecimal("-80");
            int m;
            while( (!BigDecimal.IsZero(d)) && (Math.Abs(m = n - (pi.Exponent() - d.Exponent())) > 0))
            {
                if( m < double_fig) m = double_fig;
                t   = t*m25;
                d   = BigDecimal.Div(d,ref r,t,k,m);
                k   = k + BigDecimal.Two;
                pi  = pi + d;
            }

            d = BigDecimal.One;
            k = BigDecimal.One;
            w = BigDecimal.One;
            t = new BigDecimal("956");
            while ((!BigDecimal.IsZero(d)) && ((m = n - Math.Abs(pi.Exponent() - d.Exponent()) ) > 0))
            {
                if (m < double_fig) m = double_fig;
                t   = BigDecimal.Div(null,ref r,t,m57121,n);
                d   = BigDecimal.Div(d,ref r,t,k,m);
                pi  = pi + d;
                k   = k+BigDecimal.Two;
            }
            BigDecimal.LengthRound(pi,prec);
            return pi;
        }

        /// <summary>
        /// x の正弦値（sin）を計算します。
        /// </summary>
        /// <param name="x">任意の BigDecimal </param>
        /// <param name="prec">正弦値（sin）の必要桁数</param>
        /// <returns>x の正弦値（sin）</returns>
        public static BigDecimal Sin(BigDecimal x,int prec)
        {
            int double_fig    = BigDecimal.MaxDouble.DigitCount();
            int n             = prec *2;
            BigDecimal   one  = BigDecimal.One;
            BigDecimal   two  = BigDecimal.Two;
            BigDecimal   x1   = x.Clone();
            BigDecimal   x2   = x*x;
            int  sign = 1;
            BigDecimal  y     = x.Clone();
            BigDecimal  d     = y.Clone();
            BigDecimal  i     = one.Clone();
            BigDecimal  z     = one.Clone();
            int m;
            BigDecimal r = null;
            while ((!BigDecimal.IsZero(d)) && (m = n - Math.Abs(y.Exponent() - d.Exponent())) > 0)
            {
                if (m < double_fig) m = double_fig; 
                sign = -sign;
                x1   = BigDecimal.Mul(null,x2,x1,n);
                i   += two;
                z   *= (i-one) * i;
                d    = BigDecimal.Div(d,ref r,x1,z,m);
                if(sign>0) y += d;
                else       y -= d;
            }
            BigDecimal.LengthRound(y, prec);
            return y;
        }

        /// <summary>
        /// x の余弦値（cos）を計算します。
        /// </summary>
        /// <param name="x">任意の BigDecimal </param>
        /// <param name="prec">余弦値（cos）の必要桁数</param>
        /// <returns>x の余弦値（sin）</returns>
        public static BigDecimal Cos(BigDecimal x,int prec)
        {
            int double_fig    = BigDecimal.MaxDouble.DigitCount();

            int n    = prec *2;
            BigDecimal one  = BigDecimal.One;
            BigDecimal two  = BigDecimal.Two;
            BigDecimal x1   = one.Clone();
            BigDecimal x2   = x*x;
            int        sign = 1;
            BigDecimal y    = one.Clone();
            BigDecimal d    = one.Clone();
            BigDecimal i    = BigDecimal.Zero;
            BigDecimal z    = one.Clone();
            int m = 0;
            BigDecimal r = null;
            while ( (!BigDecimal.IsZero(d)) && (m = n - Math.Abs(y.Exponent() - d.Exponent())) > 0)
            {
                if (m < double_fig) m = double_fig;
                sign = -sign;
                x1  = BigDecimal.Mul(null,x2,x1,n);
                i  += two;
                z  *= (i-one) * i;
                d    = BigDecimal.Div(d,ref r,x1,z,m);
                if(sign>0) y += d;
                else       y -= d;
            }
            BigDecimal.LengthRound(y, prec);
            return y;
        }

        /// <summary>
        /// x の逆正接関数（arctangent）です。
        /// </summary>
        /// <param name="x">任意の BigDecimal 値（x &gt; 1. である必要があります。また |x| が 0.9999 等と 1 に近いと極端に収束が遅くなります。）</param>
        /// <param name="prec">結果の必要有効桁数</param>
        /// <returns>逆正接関数値</returns>
        public static BigDecimal Atan(BigDecimal x,int prec)
        {
            int double_fig = BigDecimal.MaxDouble.DigitCount();
            int n          = prec *2;
            
            BigDecimal two = BigDecimal.Two;
            BigDecimal y = x.Clone();
            BigDecimal d = x.Clone();
            BigDecimal t = x.Clone();
            BigDecimal r = BigDecimal.Three;
            BigDecimal x2 = x*x;
            int m=0;
            BigDecimal res = null;

            while ((!BigDecimal.IsZero(d)) && (m = n - (y.Exponent() - d.Exponent()) ) > 0)
            {
                if (m < double_fig) m = double_fig; 
                t = BigDecimal.Mul(null,t,x2,n);
                BigDecimal.Negate(t);
                d = BigDecimal.Div(d,ref res,t,r,m);
                y += d;
                r += two;
            }
            BigDecimal.LengthRound(y, prec);
            return y;
        }

        /// <summary>
        /// 自然対数の底 e を prec 桁まで計算します。
        /// </summary>
        /// <param name="prec">必要有効桁数</param>
        /// <returns>自然対数の底</returns>
        public static BigDecimal E(int prec)
        {
            int double_fig    = BigDecimal.MaxDouble.DigitCount();
            int n    = prec*2;
            BigDecimal one  = BigDecimal.One;
            BigDecimal y    = BigDecimal.One;
            BigDecimal d    = BigDecimal.One;
            BigDecimal z    = BigDecimal.One;
            BigDecimal i    = BigDecimal.Zero;
            BigDecimal r = null;
            int m = 0;
            while ((!BigDecimal.IsZero(d)) && (m = n - Math.Abs(y.Exponent() - d.Exponent()) ) > 0)
            {
                if (m < double_fig) m = double_fig; 
                i += one;
                z *= i;
                d  = BigDecimal.Div(d,ref r,one,z,m);
                y += d;
            }
            BigDecimal.LengthRound(y, prec);
            return y;
        }

        /// <summary>
        /// BigDecimal 値 v の自然対数を計算します。
        /// </summary>
        /// <param name="v">自然対数を計算するBigDecimal 値</param>
        /// <param name="prec">自然対数の必要桁数</param>
        /// <returns>v の自然対数</returns>
        public static BigDecimal Log(BigDecimal v, int prec)
        {
            int double_fig = BigDecimal.MaxDouble.DigitCount();
            int n = prec * 2;

            BigDecimal one = BigDecimal.One;
            BigDecimal two = BigDecimal.Two;
            BigDecimal x = v.Clone();
            BigDecimal r = null;
            x = BigDecimal.Div(x, ref r, x - one, x + one, n);
            BigDecimal x2 = x * x;
            BigDecimal y = x.Clone();
            BigDecimal d = y.Clone();
            BigDecimal i = BigDecimal.One;
            int m = 0;
            while ((!BigDecimal.IsZero(d)) && (m = n - Math.Abs(y.Exponent() - d.Exponent())) > 0)
            {
                if (m < double_fig) m = double_fig;
                x = BigDecimal.Mul(null,x2,x,n);
                i += two;
                d = BigDecimal.Div(d, ref r, x, i, m);
                y += d;
            }
            y *= two;
            BigDecimal.LengthRound(y, prec);
            return y;
        }

        /// <summary>
        /// 自然対数の底(ネイピア数：e=2.71828...) の v 乗を計算する指数関数です。
        /// </summary>
        /// <param name="z">任意の　BigDecimal数</param>
        /// <param name="prec">必要な有効桁数</param>
        /// <returns>指数関数値</returns>
        public static BigDecimal Exp(BigDecimal v,int prec)
        {
            int double_fig = BigDecimal.MaxDouble.DigitCount();
            int n = prec * 2;
            
            BigDecimal x    = v.Clone();
            BigDecimal one  = BigDecimal.One;
            BigDecimal x1   = BigDecimal.One;
            BigDecimal y    = BigDecimal.One;
            BigDecimal d    = BigDecimal.One;
            BigDecimal z    = BigDecimal.One;
            BigDecimal i    = BigDecimal.Zero;
            BigDecimal r    = null;
            int m = 0;

            while ((!BigDecimal.IsZero(d)) && (m = n - Math.Abs(y.Exponent() - d.Exponent()) ) > 0)
            {
                if (m < double_fig) m = double_fig; 
                x1  = BigDecimal.Mul(null,x1,x,n);
                i += one;
                z *= i;
                d  = BigDecimal.Div(d,ref r, x1,z,m);
                y += d;
            }
            BigDecimal.LengthRound(y, prec);
            return y;
        }

    }
}
