﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Numerics;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;

using Mol;

namespace Examples
{
    /// <summary>
    /// LasyDFT オブジェクトが扱うフーリェ計算用インターフェース。
    /// DftFouble や DftComplex の各メソッド（プロパティ）と同じものを実装する必要があります。<br/>
    /// ※LasyDFT がデータ値等必要なときに呼び出します。インデックスは _Array.IndexBase に依存します。
    /// </summary>
    public interface IDFT
    {
        /// <summary>
        /// フーリェ変換用スケール
        /// </summary>
        double ScaleForward  { get; set; }
        /// <summary>
        /// フーリェ逆変換用スケール
        /// </summary>
        double ScaleBackward { get; set; }
        /// <summary>
        /// 元データの取得
        /// </summary>
        /// <param name="ix">取得する元データ配列のインデックス</param>
        /// <returns>元データ値（double値なら虚数値をゼロにします）</returns>
        Complex GetDataValue (int[] ix);
        /// <summary>
        /// 元データの設定
        /// </summary>
        /// <param name="v">設定値</param>
        /// <param name="ix">設定する元データ配列のインデックス</param>
        void SetDataValue(Complex v, int[] ix);
        /// <summary>
        /// 変換データの取得
        /// </summary>
        /// <param name="ix">取得する変換データ配列のインデックス</param>
        /// <returns>変換データ値（double値なら虚数値をゼロにします）</returns>
        Complex GetDftValue(int[] ix);
        /// <summary>
        /// 変換データの設定
        /// </summary>
        /// <param name="v">設定値</param>
        /// <param name="ix">設定する変換データ配列のインデックス</param>
        void SetDftValue(Complex v, int[] ix);
        /// <summary>
        /// 元（変換）データの次元（とサイズ）
        /// </summary>
        int[] Dimension { get; }
    }

    /// <summary>
    /// 離散フーリェ変換を定義式通りに計算します。
    /// FFT計算の確認や速度比較に利用できます。
    /// </summary>
    public class LasyDFT
    {
        /// <summary>
        /// 虚数単位
        /// </summary>
        private Complex iUnit = Complex.ImaginaryOne;

        /// <summary>
        /// データの次元（サイズ）
        /// </summary>
        private int[] N = null;

        /// <summary>
        /// 内部処理用の配列インデックス
        /// </summary>
        private int[] J = null;

        /// <summary>
        /// LasyDFT が利用する IDFT インターフェースを備えた（データ供給用の）計算対象オブジェクト。
        /// </summary>
        public IDFT Dft = null;

        /// <summary>
        /// デフォルトのコンストラクター
        /// </summary>
        /// <param name="dft">LasyDFT が利用する IDFT インターフェースを備えたデータ供給用計算対象オブジェクト。</param>
        public LasyDFT(IDFT dft)
        {
            Dft = dft;
            N = dft.Dimension;
            J = new int[N.Length];
        }

        /// <summary>
        /// インデックス K で指定された要素の DFTデータ値（変換データ）を元データから計算します。
        /// 計算は定義通りに、そのまま計算します。
        /// </summary>
        /// <param name="K">領域のインデックス配列</param>
        /// <returns>DFTデータ値</returns>
        public Complex ValueForward(params int[] K)
        {
            return Dft.ScaleForward * DFT(K,true);
        }

        /// <summary>
        /// インデックス K で指定された要素の元データ値を変換データから計算します。
        /// 計算は定義通りに、そのまま計算します。
        /// </summary>
        /// <param name="K">領域のインデックス配列</param>
        /// <returns>元データ値</returns>
        public Complex ValueBackward(params int[] K)
        {
            return Dft.ScaleBackward * DFT( K, false);
        }

        /// <summary>
        /// スケール無しで１要素のフーリェ変換（逆変換）を計算します。
        /// </summary>
        /// <param name="K">要素のインデックス配列</param>
        /// <param name="foward">true ならフーリェ変換、false ならフーリェ逆変換</param>
        /// <returns>計算値</returns>
        private Complex DFT(int [] K,bool foward)
        {
            Complex s = new Complex(0, 0);
            int LB = (int)_Array.IndexBase;
            for (int i = 0; i < J.Length; ++i) J[i] = LB;
        
         More:
            for (int j = 0; j < N[0]; ++j)
            {
                J[0] = j+LB;
                if (foward) s += Dft.GetDataValue(J) * Complex.Exp(-ExpPart(J, K)); // Forward 変換
                else        s += Dft.GetDftValue (J) * Complex.Exp( ExpPart(J, K)); // Backward 変換
            }

            for (int i = 1; i < N.Length; ++i)
            {
                if (++J[i] - LB < N[i]) goto More;
                J[i] = LB;
            }
            return s;
        }

        //
        // 2πiΣJ[l]*K[l]/N[l] の計算
        //
        private Complex ExpPart(int[] J, int[] K)
        {
            double s = 0;
            double ix = (double)_Array.IndexBase;
            for (int l = 0; l < J.Length; ++l)
            {
                s += ((J[l]-ix)*(K[l]-ix))/(double)N[l];
            }
            return 2.0 * T.PI * iUnit * s; // 2πi * (ΣJ[l]*K[l]/N[l])
        }
    }

    /// <summary>
    /// LasyDFT を使用するテストクラス
    /// 扱うデータは１次元 実数
    /// </summary>
    public class TestDFT : IDFT
    {
        private Complex[] m_dataVal    = null;
        private Complex[] m_dftVal     = null;
        private double m_ScaleForward  = 0.0;
        private double m_ScaleBackward = 0.0;

        DftComplex m_dft = null;
        LasyDFT    m_ldf = null;

        /// <summary>
        /// データの次元
        /// </summary>
        public int[] Dimension
        {
            get { return m_dft.Dimension; }
        }

        /// <summary>
        /// 前進変換のスケール因子
        /// </summary>
        public double ScaleForward
        {
            get
            {
                if (m_ScaleForward == 0.0) m_ScaleForward = m_dft.ScaleForward;
                return m_ScaleForward;
            }
            set
            {
                m_dft.ScaleForward = value;
                m_ScaleForward     = value;
            }
        }

        /// <summary>
        /// 後退変換のスケール因子
        /// </summary>
        public double ScaleBackward
        {
            get
            {
                if (m_ScaleBackward == 0.0) m_ScaleBackward = m_dft.ScaleBackward;
                return m_ScaleBackward;
            }
            set
            {
                m_dft.ScaleBackward = value;
                m_ScaleBackward = value;
            }
        }

        /// <summary>
        /// _DFT と LasyDFT を用いた離散フーリェ変換のテストクラス。
        /// </summary>
        /// <param name="M">データサイズ</param>
        public TestDFT(int M)
        {
            int LB = (int)_Array.IndexBase;
            m_dft = new DftComplex(1,M);
            m_ldf = new LasyDFT(this);

            m_dataVal = new Complex[M];
            m_dftVal  = new Complex[M];

            // データの作成（値に意味は無し）
            double s = -1.0;
            for (int i = LB; i < M+LB; ++i)
            {
                s = -s;
                m_dataVal[i - LB] = new Complex(s * i,-s*0.3); // 値に特に意味はありません。
                m_dft.SetDataValue(m_dataVal[i - LB], i);
            }
        }

        /// <summary>
        ///　前進変換　DftComplex と LasyDFT の処理時間を返す。 
        /// </summary>
        /// <returns>[DftComplex の前進変換時間 , LasyDFT の前進変換時間]（mS）</returns>
        public double[] ComputeForward()
        {
            int LB = (int)_Array.IndexBase;
            DateTime t0 = DateTime.Now;
            m_dft.ComputeForward();
            DateTime t1 = DateTime.Now;
            /* m_ldf (LasyDFT) の計算が必要なら以下のコメントを外す。
            for (int i = LB; i < m_dataVal.Length+LB; ++i)
            {
                m_dftVal[i-LB] = m_ldf.ValueForward(i);
            }
            */
            DateTime t2 = DateTime.Now;
            TimeSpan ts0 = t1 - t0;
            TimeSpan ts1 = t2 - t1;
            // CheckValues(); // 単純な確認（LasyDFTとの値を比較する場合）
            return new double[] { ts0.TotalMilliseconds, ts1.TotalMilliseconds };
        }

        /// <summary>
        ///　後退変換　DftComplex と LasyDFT の処理時間を返す。 
        /// </summary>
        /// <returns>[DftComplex の後退変換時間 , LasyDFT の後退変換時間]（mS）</returns>
        public double[] ComputeBackward()
        {
            int LB = (int)_Array.IndexBase;
            DateTime t0 = DateTime.Now;
            m_dft.ComputeBackward();
            DateTime t1 = DateTime.Now;
            for (int i = LB; i < m_dftVal.Length+LB; ++i)
            {
                m_dftVal[i-LB] = m_ldf.ValueBackward(i);
            }
            DateTime t2 = DateTime.Now;
            TimeSpan ts0 = t1-t0;
            TimeSpan ts1 = t2-t1;
            double s = ts1.TotalSeconds - ts0.TotalSeconds;
            CheckValues(); // 単純な確認
            return new double[] { ts0.TotalMilliseconds, ts1.TotalMilliseconds };
        }

        /// <summary>
        /// 確認のため _DFT と LasyDFT の結果を比較する。
        /// </summary>
        public void CheckValues()
        {
            int LB = (int)_Array.IndexBase;
            for (int i = LB; i < m_dataVal.Length+LB; ++i)
            {
                T.EQ(GetDataValue(i), m_dft.GetDataValue(i),0.00001);
                T.EQ(GetDftValue(i), m_dft.GetDftValue(i), 0.00001);
            }
        }

        /// <summary>
        /// LasyDFT に元データを渡すメソッド（_Array.IndexBase依存）
        /// </summary>
        /// <param name="ix">データのインデックス（_Array.IndexBase依存）</param>
        /// <returns>データ値</returns>
        public Complex GetDataValue(params int[] ix)
        {
            return m_dataVal[ix[0] - (int)_Array.IndexBase];
        }

        /// <summary>
        /// LasyDFT が元データを設定するメソッド（_Array.IndexBase依存）
        /// </summary>
        /// <param name="ix">データのインデックス（_Array.IndexBase依存）</param>
        /// <param name="ix">設定データ値</param>
        public void SetDataValue(Complex v, params int[] ix)
        {
            m_dataVal[ix[0] - (int)_Array.IndexBase] = v.Real;
        }

        /// <summary>
        /// LasyDFT に変換データを渡すメソッド（_Array.IndexBase依存）
        /// </summary>
        /// <param name="ix">データのインデックス（_Array.IndexBase依存）</param>
        /// <returns>データ値</returns>
        public Complex GetDftValue(params int[] ix)
        {
            return m_dftVal[ix[0] - (int)_Array.IndexBase];
        }

        /// <summary>
        /// LasyDFT が変換データを設定するメソッド（_Array.IndexBase依存）
        /// </summary>
        /// <param name="ix">データのインデックス（_Array.IndexBase依存）</param>
        /// <param name="ix">設定データ値</param>
        public void SetDftValue(Complex v, params int[] ix)
        {
            m_dftVal[ix[0] - (int)_Array.IndexBase] = v;
        }
    }
}
