﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Numerics;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;

using Mol;

namespace Examples
{
    /// <summary>
    /// 基本統計 Stat のテストプログラム
    /// </summary>
    public class TestStat
    {
        /// <summary>
        /// 行列 data の分散共分散行列を（定義通りに）計算します。
        /// ※ Stat の結果確認用です。
        /// </summary>
        /// <param name="data">データ行列</param>
        /// <returns>分散共分散行列</returns>
        public MatrixDenseSymmetricDouble LasyCov(MatrixDenseGeneralDouble data)
        {
            int P = data.ColumnCount;
            int N = data.RowCount;
            MatrixDenseSymmetricDouble C = new MatrixDenseSymmetricDouble(P);
            VectorDenseDouble M  = new VectorDenseDouble(P);
            VectorDenseDouble S2 = new VectorDenseDouble(P);
            for (int j = 1; j <= P; ++j)
            {
                for (int i = 1; i <= N; ++i)
                {
                    M[j] += data[i, j];
                    S2[j] += data[i, j] * data[i, j];
                }
            }
            for (int j = 1; j <= P; ++j) M[j] /= N; // 平均
            for (int i = 1; i <= P; ++i)
            {
                for (int j = i; j <= P; ++j)
                {
                    for (int k = 1; k <= N; ++k)
                    {
                        C[i, j] += (data[k, i] - M[i]) * (data[k, j] - M[j]);
                    }
                }
            }
            for (int i = 1; i <= P; ++i)
            {
                for (int j = i; j <= P; ++j)
                {
                    C[i, j] /= (N-1);  // 共分散
                }
            }
            return C;
        }

        /// <summary>
        /// 分散共分散行列から相関行列を（定義通りに）計算します。
        /// </summary>
        /// <param name="cov">分散共分散行列</param>
        /// <returns>相関行列</returns>
        public MatrixDenseSymmetricDouble LasyCor(MatrixDenseSymmetricDouble cov)
        {
            int P = cov.ColumnCount;
            MatrixDenseSymmetricDouble C = new MatrixDenseSymmetricDouble(P);
            for (int i = 1; i <= P; ++i)
            {
                for (int j = i; j <= P; ++j)
                {
                    C[i, j] = cov[i, j] / Math.Sqrt(cov[i, i] * cov[j, j]);
                }
            }
            return C;
        }

        /// <summary>
        /// 分割処理のテスト。
        /// 最初に大きな DataMatrix に乱数をセットして基本統計量（stat1）を計算しておきます。
        /// その後、大きなデータ行列を行単位に分割して「分割処理」を実行（stat2）して結果を比較します。
        /// </summary>
        public void Test1()
        {
            int N = 20000; // データレコード数
            int P = 4;     // 次元
            int D = 15;    // 分割数

            T.P("TestStat.Test1(): Started");
            
            // データ生成用作業ベクトル
            VectorDenseDouble rec = new VectorDenseDouble(N);

            // 全データ
            MatrixDenseGeneralDouble data1 = new MatrixDenseGeneralDouble(N, P);
            // 全データの設定（乱数）            
            for (int i = 1; i <= P; ++i)
            {
                RandomNumber rnd = new RandomNumber(RandomNumber.GENERATOR.MT2203, 7777 + (uint)i);
                // 統計データ用に乱数を発生させ（平均値0.0、標準偏差 i）、行列の i 列にセットする。
                data1.Columns[i] = rnd.GenerateGaussian(rec, 0.0, (double)i, RandomNumber.GAUSSIAN_METHOD.BOXMULLER);
                _Mol.EQ(data1.Columns[i], rec);
            }

            //
            // 基本統計量の計算
            //
            VectorDenseDouble QuantileOrder = new VectorDenseDouble(3); // 分位率
            QuantileOrder[1] = 0.25;
            QuantileOrder[2] = 0.5;
            QuantileOrder[3] = 0.75;

            //
            // ========= 一括計算 ===========
            //
            Stat stat1 = new Stat(data1);
            stat1.Computed = Stat.PROPERTIES.ALL; // 全てのプロパティを一括計算。
            stat1.QuantileOrder = QuantileOrder;  // 分位率を設定して対応する分位値行列を計算する。
            stat1.ComputeProperties();
            PrintStat("一括計算の結果", stat1);
            MatrixDenseGeneralDouble QM = stat1.ComputeQuantileMatrix(null, null, stat1.QuantileOrder);
            T.P("分位値行列（計算値）", QM);

            // 直接計算との比較
            T.EQ(LasyCor(LasyCov(data1)),stat1.CorrelationMatrix);

            //
            // ====== 分割計算 =======
            //
            int M = N / D; // 全体データ行列(data1)を D 分割して小データで繰り返して計算する。
            MatrixDenseGeneralDouble data2 = new MatrixDenseGeneralDouble(M,P);
            Stat stat2 = new Stat(data2);
            stat2.Computed = Stat.PROPERTIES.ALL; // 全てのプロパティを計算。
            stat2.QuantileOrder = QuantileOrder;  // 同じ分位率を設定

            data2 = new MatrixDenseGeneralDouble(M, P);
            int ni = (N + M - 1) / M;
            int I = 1;
            for(int i=0;i<ni;++i)
            {
                // 全体データ(data1)の分割部分を小データ(date2)にコピー
                stat2.DataMatrix = _Matrix.SubMatrix(data2, data1, I, M, 1, P);
                stat2.ComputeProperties(i);
                I += M;
            }
            PrintStat("分割計算の結果", stat2);
            T.EQ(stat1.Min, stat2.Min);
            T.EQ(stat1.Max, stat2.Max);
            T.EQ(stat1.Mean, stat2.Mean);
            T.EQ(stat1.Variance, stat2.Variance);
            T.EQ(stat1.CovarianceMatrix, stat2.CovarianceMatrix);
            T.EQ(stat1.CorrelationMatrix, stat2.CorrelationMatrix);
            T.EQ(stat1.QuantileMatrix, stat2.QuantileMatrix);
            T.P("パラメータ化された相関行列",stat1.ParameterizeCorrelationMatrix(null));

            // 外れ値の検出
            // 棄却領域
            double critical_level = 0.05;
            VectorDenseDouble critical_min = new VectorDenseDouble(P);
            VectorDenseDouble critical_max = new VectorDenseDouble(P);
            for (int i = 1; i <= P; ++i)
            {
                critical_min[i] = critical_level;
                critical_max[i] = 1.0 - critical_level;
            }
            _VML.CdfNormInv(critical_min);
            _VML.CdfNormInv(critical_max);
            for (int i = 1; i <= P; ++i)
            {
                critical_min[i] *= (double)i;
                critical_max[i] *= (double)i;
            }

            T.P("左棄却領域", critical_min);
            T.P("右棄却領域", critical_max);

            VectorDenseDouble out_w = stat1.DetectOutliers(null, Stat.BACON_INITILIZER.MAHALANOBIS, critical_level, 0.005);
            int no = 0;
            int nd = 0;
            VectorDenseDouble tv = new VectorDenseDouble(1);
            VectorDenseDouble v1 = new VectorDenseDouble(P);
            for (int i = 1; i <= N; ++i)
            {
                if (out_w[i] > 0) continue;
                no++;
                bool fin = false;
                for (int j = 1; j <= P; ++j)
                {
                    // 外れ値が正規分布の棄却領域内であることを確認（ただし誤差あり）
                    if (data1[i, j] <= critical_min[j] || data1[i, j] >= critical_max[j])
                    {
                        nd++;
                        fin = true;
                        break;
                    }
                }
                if (!fin)
                {
                    T.P("棄却領域外の外れ値", data1.Rows[i]);
                }
            }
            T.WL("外れ値の数=" + no + " 棄却域に入った数="+nd);

            stat1.ComputeRobustStat(0, 0.05, 0.001, 20);

            T.P("ロバスト推定（平均)", stat1.Mean);
            T.P("ロバスト推定（分散共分散行列）", stat1.CovarianceMatrix);

            //
            // 欠測値の推定
            int[] IX = new int[20];
            int[] JX = new int[20];
            double[] missing_vals = new double[20];
            IX[0] = 1; JX[0] = 1;
            IX[1] = 200; JX[1] = 2;
            IX[2] = 300; JX[2] = 3;
            IX[3] = 400; JX[3] = 4;
            IX[4] = 500; JX[4] = 1;
            IX[5] = 600; JX[5] = 2;
            IX[6] = 700; JX[6] = 3;
            IX[7] = 800; JX[7] = 4;
            IX[8] = 900; JX[8] = 1;
            IX[9] = 1000; JX[9] = 2;
            IX[10] = 1100; JX[10] = 3;
            IX[11] = 1200; JX[11] = 4;
            IX[12] = 1300; JX[12] = 1;
            IX[13] = 1400; JX[13] = 2;
            IX[14] = 1500; JX[14] = 3;
            IX[15] = 1600; JX[15] = 4;
            IX[16] = 1700; JX[16] = 1;
            IX[17] = 1800; JX[17] = 2;
            IX[18] = 1900; JX[18] = 3;
            IX[19] = 2000; JX[19] = 4;
            for (int i = 0; i < IX.Length; ++i)
            {
                missing_vals[i] = data1[IX[i], JX[i]];
                data1[IX[i], JX[i]] = Stat.MissingValue;
            }
            DateTime dt = DateTime.Now;
            stat1.ReplaceMissingValues(0, 0, 0, 0);
            T.WL("欠測値の推定計算時間："+(DateTime.Now-dt));

            T.WL("欠測値の推定");
            double sm = 0.0;
            for (int i = 0; i < IX.Length; ++i)
            {
                T.WL("元の値="+T.R(missing_vals[i])+"  推定値="+T.R(data1[IX[i], JX[i]]));
                sm += (missing_vals[i] - data1[IX[i], JX[i]]) * (missing_vals[i] - data1[IX[i], JX[i]]);
            }
            T.WL("残差=" + Math.Sqrt(sm));

            stat1.Dispose();
            stat2.Dispose();
            T.P("TestStat.Test1(): End");
        }

        /// <summary>
        /// 一括計算メソッドの確認
        /// </summary>
        public void Test2()
        {
            int N = 2000;  // データレコード数
            int P = 8;     // 次元（２で割り切れる数）

            T.P("TestStat.Test2(): Started");

            // データ生成用作業ベクトル
            VectorDenseDouble rec = new VectorDenseDouble(N);

            // 全データ
            MatrixDenseGeneralDouble data1 = new MatrixDenseGeneralDouble(N, P);
            // 全データの設定（乱数）            
            for (int i = 1; i <= P; ++i)
            {
                RandomNumber rnd = new RandomNumber(RandomNumber.GENERATOR.MT2203, 7777 + (uint)i);
                // 統計データ用に乱数を発生させる（平均値0.0、標準偏差 i）。
                if(i%2==1) rnd.GenerateGaussian(rec, 0.0, (double)i, RandomNumber.GAUSSIAN_METHOD.BOXMULLER);
                else       rnd.GenerateBinomial(rec,N, 0.5); // 平均:0.5N、 分散：0.25N
                // 行列の i 列に発生させた乱数をセットする。
                data1.Columns[i] = rec;
                _Mol.EQ(data1.Columns[i], rec);
            }

            //
            // 全プロパティの一括計算
            Stat stat1 = new Stat(data1);
            stat1.Computed = Stat.PROPERTIES.ALL; // 全てのプロパティを一括計算。
            stat1.ComputeProperties();
            PrintStat("一括計算の結果", stat1);

            // 直接計算との比較
            MatrixDenseSymmetricDouble cv = LasyCov(data1);
            T.EQ(cv, stat1.CovarianceMatrix,0.00001);

            MatrixDenseSymmetricDouble cr = LasyCor(cv);
            T.EQ(cr, stat1.CorrelationMatrix,0.00001);
            T.P("パラメータ化された相関行列", stat1.ParameterizeCorrelationMatrix(null));

            // 
            // 部分分散・相関行列
            int P2 = P / 2;
            VectorDenseInt divider = new VectorDenseInt(P);

            MatrixDenseGeneralDouble Cz = new MatrixDenseGeneralDouble(P2, P2);
            MatrixDenseGeneralDouble Cy = new MatrixDenseGeneralDouble(P2, P2);
            MatrixDenseGeneralDouble Czy = new MatrixDenseGeneralDouble(P2, P2);
            MatrixDenseGeneralDouble Cyz = new MatrixDenseGeneralDouble(P2, P2);

            MatrixDenseSymmetricDouble pcov = null;
            MatrixDenseSymmetricDouble pcor = null;
            int dv = -1;
            for (int i = 1; i <= P; ++i)
            {
                divider[i] = dv;
                dv *= -1;
            }

            //
            // 定義に従って部分分散を計算する。
            MatrixDenseGeneralDouble CV =  new MatrixDenseGeneralDouble(P,P);
            CV += stat1.CovarianceMatrix; // コピー（0を足す）
            int J = 0;
            VectorDenseInt div2 = (VectorDenseInt)divider.Clone();
            for (int i = 1; i <= div2.Count; ++i) div2[i] = i*div2[i];
            // divider[] の指定に従って行を並べ替える。
            for (int j = 1; j <= P; ++j)
            {
                if (divider[j] > 0) continue;
                ++J;
                // 行と列の並べ替え
                if (j == J) continue;
                VectorDenseDouble Tv = CV.Rows[J];
                CV.Rows[J] = CV.Rows[j];
                CV.Rows[j] = Tv;
                Tv = CV.Columns[J];
                CV.Columns[J] = CV.Columns[j];
                CV.Columns[j] = Tv;
                int Tmp = div2[J];
                div2[J] = div2[j];
                div2[j] = Tmp;
            }
            // +1 部分をソートする(-1 はソートされている)
            for (int j = J+1; j <= P; ++j)
            {
                int  vmin = div2[j];
                int  jmin = j;
                for (int jj = j + 1; jj <= P; ++jj)
                {
                    if (vmin > div2[jj])
                    {
                        jmin = jj;
                        vmin = div2[jj];
                    }
                }
                if (jmin > j)
                {
                    vmin = div2[jmin];
                    div2[jmin] = div2[j];
                    div2[j] = vmin;
                    VectorDenseDouble Tv = CV.Rows[jmin];
                    CV.Rows[jmin] = CV.Rows[j];
                    CV.Rows[j] = Tv;
                    Tv = CV.Columns[jmin];
                    CV.Columns[jmin] = CV.Columns[j];
                    CV.Columns[j] = Tv;
                }
            }
            T.P("並べ替え後の列", div2);
            T.P("並べ替え後の分散共分散行列", CV);

            // Z & YZ
            for (int j = 1; j <= J; ++j)
            {
                for (int i = 1; i <= J; ++i)
                {
                    Cz[i,j] = CV[i, j];
                }
                for (int i = J+1; i <= P; ++i)
                {
                    Cyz[i-J, j] = CV[i, j];
                }
            }
            // Y & ZY
            for (int j = J+1; j <= P; ++j)
            {
                for (int i = 1; i <= J; ++i)
                {
                    Czy[i, j - J] = CV[i, j];
                }
                for (int i = J + 1; i <= P; ++i)
                {
                    Cy[i - J, j-J] = CV[i, j];
                }
            }
            T.P("Cz", Cz);
            T.P("Cy", Cy);
            T.P("Czy", Czy);
            T.EQ(Czy,Cyz.Transpose());

            LuSolver lu = LuSolver.Create(Cz);
            Cy = Cy - Cyz * lu.Solve(Czy);
            T.P("部分分散（計算値）", Cy);

            stat1.ComputePartialCovarianceCorrelationMatrix(ref pcov, ref pcor, divider);
            T.P("部分分散", pcov);
            T.P("部分相関", pcor);
            Cy = pcov - Cy;
            T.WL("部分分散の残差：　" + Cy.NormE());
            for (int i = 1; i <= P / 2; ++i)
            {
                for (int j = 1; j <= P / 2; ++j)
                {
                    Cy[i, j] = pcov[i, j] / Math.Sqrt(pcov[i, i] * pcov[j, j]);
                }
            }
            Cy = pcor - Cy;
            T.WL("部分相関の残差：　" + Cy.NormE());

            //
            // Pooled & Grouped Estimates
            VectorDenseInt pdivider = new VectorDenseInt(N);
            J = 1;
            while (J <= N)
            {
                for (int i = 0; i < P; ++i)
                {
                    pdivider[J++] = i;
                }
            }
            VectorDenseDouble            p_mean = null;
            MatrixDenseSymmetricDouble   p_cov = null;
            MatrixDenseGeneralDouble     g_means = null;
            MatrixDenseSymmetricDouble[] g_cov = null;
            stat1.ComputePooledGroupEstimates(pdivider, ref p_mean, ref p_cov, ref g_means, ref g_cov);
            T.P("Pooled mean", p_mean);
            T.P("Pooled Covariance matrix", p_cov);
            T.P("Group means", g_means);
            for (int i = 1; i <= P; ++i)
            {
                T.P("Group Covariance matrix["+i+"]", g_cov[i-1]);
            }

            T.P("TestStat.Test2(): End");
        }

        private void PrintStat(string title, Stat s)
        {
            T.P(title);
            T.P("最小", s.Min);
            T.P("最大", s.Max);
            T.P("平均", s.Mean);
            T.P("分散", s.Variance);
            T.P("分散共分散行列", s.CovarianceMatrix);
            T.P("相関行列", s.CorrelationMatrix);
            if(s.QuantileMatrix!=null) T.P("分位値行列(推定値)", s.QuantileMatrix);
        }
    }
}
